import bs
import bsUtils
import bsPowerup
import bsBomb
from bsVector import Vector
import bsAnnounce
import random
import weakref
import bsVector

_gFactory = None

# list of defined spazzes
appearances = {}

def getAppearances():
    import _bs
    disallowed = []
    if not _bs._getPurchased('santa'): disallowed.append('Santa Claus')
    if not _bs._getPurchased('frosty'): disallowed.append('Frosty')
    return [s for s in appearances.keys() if s not in disallowed]

gPowerupWearOffTime = 20000
gBasePunchPowerScale = 1.2
gBasePunchCooldown = 400

gDefaultBotColor = (0.45,0.6,0.42)
gProBotColor = (1.0,0.2,0.1)
gLameBotColor = (1,0.8,0.0)

gDefaultBotHighlight = (0.1,0.15,0.1)
gProBotHighlight = (0.6,0.1,0.05)
gLameBotHighlight = (0.6,0.4,0.15)


class PlayerSpazDeathMessage(object):
    """A PlayerSpaz has died."""
    def __init__(self,spaz,wasKilled,killerPlayer,how):
        self.spaz = spaz
        self.killed = wasKilled
        self._killerPlayer = None if killerPlayer is None else weakref.ref(killerPlayer)
        self.how = how
    def getKillerPlayer(self):
        return None if self._killerPlayer is None else self._killerPlayer()

class PlayerSpazHurtMessage(object):
    def __init__(self,spaz):
        self.spaz = spaz

class CurseExplodeMessage(object):
    pass

class AttackMessage(object):
    def __init__(self,type):
        self.type = type

class BombDeathMsg(object):
    "A bomb has died and thus can be recycled"
    pass


class Appearance(object):
    """Create and fill out one of these suckers to define a spaz appearance"""
    def __init__(self,name):

        self.name = name

        if appearances.has_key(self.name):
            raise Exception("spaz appearance name \"" + self.name + "\" already exists.")

        appearances[self.name] = self
        self.colorTexture = ""
        self.headModel = ""
        self.torsoModel = ""
        self.pelvisModel = ""
        self.upperArmModel = ""
        self.foreArmModel = ""
        self.handModel = ""
        self.upperLegModel = ""
        self.lowerLegModel = ""
        self.toesModel = ""
        self.jumpSounds = []
        self.attackSounds = []
        self.impactSounds = []
        self.deathSounds = []
        self.pickupSounds = []
        self.fallSounds = []
        self.style = 'spaz'

class PunchMessage(object):
    "Message saying we want to punch"
    pass

class PunchHitMessage(object):
    "Message saying an object was hit"
    pass

class PickupMessage(object):
    "Message saying we wanna pick something up"
    pass

class Factory(object):
    """ wraps up media an other resources used by spazzes """

    def preload(self,character):
        """Preload media that'll be needed for a given spaz type"""
        self.getMedia(character)

    def __init__(self):

        self.medImpactSounds = bs.newMediaGroup((bs.getSound("impactMedium"),
                                                 bs.getSound("impactMedium2")))
        self.hardImpactSounds = bs.newMediaGroup((bs.getSound("impactHard"),
                                                  bs.getSound("impactHard2"),
                                                  bs.getSound("impactHard3")))
        self.bigImpactSounds = bs.newMediaGroup((bs.getSound("bigImpact"),
                                                 bs.getSound("bigImpact2")))
        self.singlePlayerDeathSound = bs.getSound("playerDeath")
        self.punchSound = bs.getSound("punch01")
        
        self.punchSoundStrong = bs.newMediaGroup((bs.getSound("punchStrong01"),
                                                  bs.getSound("punchStrong02")))
        self.superPunchSound = bs.getSound("superPunch")
        
        self.popSound = bs.getSound("pop01")
        self.swishSound = bs.getSound("punchSwish")
        self.blockSound = bs.getSound("block")
        self.shatterSound = bs.getSound("shatter")
        self.splatterSound = bs.getSound("splatter")
        
        self.spazMaterial = bs._newMaterial()
        self.rollerMaterial = bs._newMaterial()
        self.punchMaterial = bs._newMaterial()
        self.pickupMaterial = bs._newMaterial()
        self.curseMaterial = bs._newMaterial()

        # send footing messages to spazzes so they know when they're on solid ground
        # eww this should really just be built into the spaz node
        bs._materialAddAction(self.rollerMaterial,
                             conditions=("dstIsMaterial",bs.footingMaterial),
                             actions=(("message","self","atConnect","footing",1),
                                      ("message","self","atDisconnect","footing",-1)))

        bs._materialAddAction(self.spazMaterial,
                             conditions=("dstIsMaterial",bs.footingMaterial),
                             actions=(("message","self","atConnect","footing",1),
                                      ("message","self","atDisconnect","footing",-1)))
        # punches
        bs._materialAddAction(self.punchMaterial,
                             conditions=("srcDstDiffNode",),
                             actions=(("partMod","collide",True),
                                      ("partMod","physical",False),
                                      ("message","self","atConnect",PunchHitMessage())))
        # pickups
        bs._materialAddAction(self.pickupMaterial,
                             conditions=(("srcDstDiffNode",),'and',("dstIsMaterial",bs.objectMaterial)),
                             actions=(("partMod","collide",True),
                                      ("partMod","physical",False),
                                      ("message","self","atConnect",PickupMessage())))
        # curse
        bs._materialAddAction(self.curseMaterial,
                             conditions=(("srcDstDiffNode",),'and',("dstIsMaterial",bs.playerMaterial)),
                             actions=(("message","self","atConnect",CurseExplodeMessage()),))


        self.footImpactSounds = bs.newMediaGroup(( bs.getSound("footImpact01"),
                                                   bs.getSound("footImpact02"),
                                                   bs.getSound("footImpact03") ))

        self.footSkidSound = bs.getSound("skid01")
        self.footRollSound = bs.getSound("scamper01")

        bs._materialAddAction(self.rollerMaterial,
                             conditions=("dstIsMaterial",bs.footingMaterial),
                             actions=(("impactSound",self.footImpactSounds,1,0.2),
                                      ("skidSound",self.footSkidSound,20,0.3),
                                      #("skidSound",self.footSkidSound,0.1,50.0),
                                      ("rollSound",self.footRollSound,20,3.0)))

        self.skidSound = bs.getSound("gravelSkid")

        bs._materialAddAction(self.spazMaterial,
                             conditions=("dstIsMaterial",bs.footingMaterial),
                             actions=(("impactSound",self.footImpactSounds,20,6),
                                      ("skidSound",self.skidSound,2.0,1),
                                      ("rollSound",self.skidSound,2.0,1)))

        self.shieldUpSound = bs.getSound("shieldUp")
        self.shieldDownSound = bs.getSound("shieldDown")
        self.shieldHitSound = bs.getSound("shieldHit")

        # we dont want to collide with stuff we're initially overlapping
        # (unless its marked with a special region material)
        bs._materialAddAction(self.spazMaterial,
                             conditions=( (("srcYoungerThan", 51),'and',('srcDstDiffNode',)),'and',('dstNotMaterial',bs.regionMaterial)),
                             actions=( ("nodeMod","collide",False)))
        self.spazMedia = {}

    def getStyle(self,character):
        return appearances[character].style
        
    def getMedia(self,character):

        t = appearances[character]
        if not self.spazMedia.has_key(character):
            m = self.spazMedia[character] = \
                ([0] # placeholder for texture
                 + [0] # placeholder for tint texture
                 + [bs.getModel(t.headModel)]
                 + [bs.getModel(t.torsoModel)]
                 + [bs.getModel(t.pelvisModel)]
                 + [bs.getModel(t.upperArmModel)]
                 + [bs.getModel(t.foreArmModel)]
                 + [bs.getModel(t.handModel)]
                 + [bs.getModel(t.upperLegModel)]
                 + [bs.getModel(t.lowerLegModel)]
                 + [bs.getModel(t.toesModel)]
                 + [len(t.jumpSounds)] + [bs.getSound(s) for s in t.jumpSounds]
                 + [len(t.attackSounds)] + [bs.getSound(s) for s in t.attackSounds]
                 + [len(t.impactSounds)] + [bs.getSound(s) for s in t.impactSounds]
                 + [len(t.deathSounds)] + [bs.getSound(s) for s in t.deathSounds]
                 + [len(t.pickupSounds)] + [bs.getSound(s) for s in t.pickupSounds]
                 + [len(t.fallSounds)] + [bs.getSound(s) for s in t.fallSounds])
        else:
            m = self.spazMedia[character]
        
        # update the color tex for this character
        m[0] = bs.getTexture(t.colorTexture)
        m[1] = bs.getTexture(t.colorMaskTexture)
        return m


def getFactory():
    """ return the global factory object, initializing it if necessary """
    global _gFactory
    if _gFactory == None:
        _gFactory = Factory()
    return _gFactory


class Spaz(bs.Actor):

    pointsMult = 1
    curseTime = 5000

    defaultBombCount = 1
    defaultBombType = 'normal'
    defaultBoxingGloves = False
    defaultShields = False

    def __init__(self,activity,retain,color=(1,1,1),highlight=(0.5,0.5,0.5),character="Spaz",source=None,fly=False,startInvincible=True,hockey=False,canAcceptPowerups=True,powerupsExpire=False):
        bs.Actor.__init__(self,activity,retain)

        factory = getFactory()

        self.source = source
        self.dead = False
        self.punchPowerScale = gBasePunchPowerScale
        self.fly = fly
        self.hockey = hockey
        self._dontPickThisUp = None
        self.playBigDeathSound = False
        self._punchedNodes = set()
        self._cursed = False
        self._connectedToPlayerID = None

        materials = [factory.spazMaterial,bs.objectMaterial,bs.playerMaterial]
        rollerMaterials = [factory.rollerMaterial,bs.playerMaterial]

        if canAcceptPowerups:
            materials.append(bsPowerup.getFactory().powerupAcceptMaterial)
            rollerMaterials.append(bsPowerup.getFactory().powerupAcceptMaterial)

        self.node = bs.newNode("spaz",
                               owner=activity,
                               attrs={'color':color,
                                      'highlight':highlight,
                                      'media':factory.getMedia(character),
                                      'style':factory.getStyle(character),
                                      'fly':self.fly,
                                      'hockey':self.hockey,
                                      'materials':materials,
                                      'rollerMaterials':rollerMaterials,
                                      'punchMaterials':(factory.punchMaterial,bs.attackMaterial),
                                      'pickupMaterials':(factory.pickupMaterial,bs.pickupMaterial),
                                      'invincible':startInvincible,
                                      'sourcePlayer':source},
                               delegate=self)
        self.shield = None

        if startInvincible:
            def _safeSetAttr(node,attr,val):
                if node.exists(): setattr(node,attr,val)
            bs.gameTimer(1000,bs.WeakCall(_safeSetAttr,self.node,'invincible',False))

        self.hitPoints = 1000
        self.hitPointsMax = 1000
        self.bombCount = self.defaultBombCount
        self.maxBombCount = self.defaultBombCount
        self.setSpeed(1.0)
        self.bombTypeDefault = self.defaultBombType
        self.bombType = self.bombTypeDefault
        self.landMineCount = 0
        self.blastRadius = 2.0
        self.holdingNode = None
        self.powerupsExpire = powerupsExpire

        self.punchCooldown = gBasePunchCooldown
        self.hasBoxingGloves = False
        if self.defaultBoxingGloves: self.getBoxingGloves()
        self.lastPunchTime = 0
        self.frozen = False
        self.shattered = False

        self._lastHitTime = None
        self._numTimesHit = 0
        self._bombHeld = False
        if self.defaultShields: self.shieldsUp()

        self.dropBombCallback = None
        self.punchCallback = None
        self.pickUpPowerupCallback = None


        # demonstration of our weird node-update-order issue..
        if 0:
            scoreText = bs.newNode('text',
                                          owner=self.node,
                                          attrs={'text':'TESTING',
                                                 'inWorld':True,
                                                 'color':(1,0,0,1),
                                                 'scale':[0.02],
                                                 'hAlign':'center'},
                                          name='spaz score txt')
            m = bs.newNode('math',owner=self.node,attrs={'input1':(0,1.7,0),'operation':'add'})
            bs.connectAttr(self.node,'torsoTranslate',m,'input2')
            bs.connectAttr(m,'output',scoreText,'translate')

            m = bs.newNode('math',owner=self.node,attrs={'input1':(0,1.4,0),'operation':'add'})
            bs.connectAttr(self.node,'torsoTranslate',m,'input2')
            scoreText = bs.newNode('text',
                                          owner=self.node,
                                          attrs={'text':'TESTING',
                                                 'inWorld':True,
                                                 'color':(1,0,0,1),
                                                 'scale':[0.02],
                                                 'hAlign':'center'},
                                          name='spaz score txt')
            bs.connectAttr(m,'output',scoreText,'translate')


    def setSpeed(self,speed):
        self.speed = speed
        self.node.speed = self.speed

    def getActivity(self):
        return self._activity()

    def exists(self):
        return self.node.exists()

    def onJumpRelease(self):
        bs.nodeMessage(self.node,"jumpRelease")

    def _hideScoreText(self):
        if self._scoreText.exists():
            bsUtils.animate(self._scoreText,'scale',{0:self._scoreText.scale[0],200:0})

    def setScoreText(self,t,color=(1,1,0.4),flash=False):
        """ utility func to show a message over our spaz; handy for scores """
        colorFin = bsUtils.getSafeDisplayColor(color)[:3]
        if not self.node.exists(): return
        try: exists = self._scoreText.exists()
        except Exception: exists = False
        if not exists:
            startScale = 0.0
            m = bs.newNode('math',owner=self.node,attrs={'input1':(0,1.4,0),'operation':'add'})
            bs.connectAttr(self.node,'torsoTranslate',m,'input2')
            self._scoreText = bs.newNode('text',
                                          owner=self.node,
                                          attrs={'text':t,
                                                 'inWorld':True,
                                                 'color':colorFin,
                                                 'scale':[0.02],
                                                 'hAlign':'center'},
                                          name='spaz score txt')
            bs.connectAttr(m,'output',self._scoreText,'translate')
        else:
            self._scoreText.color = colorFin
            startScale = self._scoreText.scale[0]
            self._scoreText.text = t
        if flash:
            combine = bs.newNode("combine",owner=self._scoreText,attrs={'size':3})
            sc = 1.8
            offs = 0.5
            t = 300
            for i in range(3):
                c1 = offs+sc*colorFin[i]
                c2 = colorFin[i]
                bsUtils.animate(combine,'input'+str(i),{0.5*t:c2,
                                                        0.75*t:c1,
                                                        1.0*t:c2})
            bs.connectAttr(combine,'output',self._scoreText,'color')
            
        bsUtils.animate(self._scoreText,'scale',{0:startScale,200:0.02})
        self._scoreTextHideTimer = bs.Timer(1000,bs.WeakCall(self._hideScoreText))

    def onJumpPress(self):
        bs.nodeMessage(self.node,"jump")

    def onPickUpPress(self):
        bs.nodeMessage(self.node,'pickUp')

    def onPickUpRelease(self):
        bs.nodeMessage(self.node,'pickUpRelease')

    def onHoldPositionPress(self):
        bs.nodeMessage(self.node,'holdPosition')

    def onHoldPositionRelease(self):
        bs.nodeMessage(self.node,'holdPositionRelease')

    def onPunchPress(self):
        bs.nodeMessage(self.node,AttackMessage('punch'))
        
    def onPunchRelease(self):
        bs.nodeMessage(self.node,'punchRelease')

    def onBombPress(self):
        if self.dead or self.frozen: return
        if self.node.knockout > 0.0: return
        if self.holdingNode == None:
            self.dropBomb()
            self.node.bombButtonHeld = True
        else: bs.nodeMessage(self.node,'pickUp') # this throws

    def onBombRelease(self):
        if self.node.exists(): self.node.bombButtonHeld = False

    def onRunPress(self):
        bs.nodeMessage(self.node,'run')

    def onRunRelease(self):
        bs.nodeMessage(self.node,'runRelease')

    def onFlyPress(self):
        bs.nodeMessage(self.node,'fly')

    def onFlyRelease(self):
        bs.nodeMessage(self.node,'flyRelease')

    def onMoveUpDown(self):
        bs.nodeMessage(self.node,"moveUpDown",bs.getInputArgument())

    def onMoveLeftRight(self):
        bs.nodeMessage(self.node,"moveLeftRight",bs.getInputArgument())


    def onPunched(self,damage):
        pass

    def getDeathPoints(self,how):
        numHits = float(max(1,self._numTimesHit))
        # base points is simply 10 for 1-hit-kills and 5 otherwise
        importance = 2 if numHits < 2 else 1
        return ((10 if numHits < 2 else 5) * self.pointsMult,importance)

    def disconnectFromPlayer(self):
        if self._connectedToPlayerID is not None:
            bs.resetPlayerInput(self._connectedToPlayerID)
            self._connectedToPlayerID = None
            # send releases for anything in case its held..
            # FIXME - send zeroed moves (cant currently do this due to bs.getInputArgument)
            self.onHoldPositionRelease()
            self.onJumpRelease()
            self.onPickUpRelease()
            self.onPunchRelease()
            self.onBombRelease()
            self.onRunRelease()
            self.onFlyRelease()
        else: print 'WARNING: disconnectFromPlayer() called for non-connected player'

    def connectToPlayer(self,playerID,enableJump=True,enablePunch=True,enablePickUp=True,enableBomb=True,enableRun=True,enableFly=True):

        # reset any currently connected player and/or the player we're now wiring up
        if self._connectedToPlayerID is not None:
            if playerID != self._connectedToPlayerID: bs.resetPlayerInput(playerID)
            self.disconnectFromPlayer()
        else: bs.resetPlayerInput(playerID)

        bs.assignPlayerInput(playerID,"upDown",bs.WeakCall(self.onMoveUpDown))
        bs.assignPlayerInput(playerID,"leftRight",bs.WeakCall(self.onMoveLeftRight))
        bs.assignPlayerInput(playerID,"holdPositionPress",bs.WeakCall(self.onHoldPositionPress))
        bs.assignPlayerInput(playerID,"holdPositionRelease",bs.WeakCall(self.onHoldPositionRelease))

        if enableJump:
            bs.assignPlayerInput(playerID,"jumpPress",bs.WeakCall(self.onJumpPress))
            bs.assignPlayerInput(playerID,"jumpRelease",bs.WeakCall(self.onJumpRelease))
        if enablePickUp:
            bs.assignPlayerInput(playerID,"pickUpPress",bs.WeakCall(self.onPickUpPress))
            bs.assignPlayerInput(playerID,"pickUpRelease",bs.WeakCall(self.onPickUpRelease))
        if enablePunch:
            bs.assignPlayerInput(playerID,"punchPress",bs.WeakCall(self.onPunchPress))
            bs.assignPlayerInput(playerID,"punchRelease",bs.WeakCall(self.onPunchRelease))
        if enableBomb:
            bs.assignPlayerInput(playerID,"bombPress",bs.WeakCall(self.onBombPress))
            bs.assignPlayerInput(playerID,"bombRelease",bs.WeakCall(self.onBombRelease))
        if enableRun:
            bs.assignPlayerInput(playerID,'runPress',bs.WeakCall(self.onRunPress))
            bs.assignPlayerInput(playerID,'runRelease',bs.WeakCall(self.onRunRelease))
        if enableFly:
            bs.assignPlayerInput(playerID,'flyPress',bs.WeakCall(self.onFlyPress))
            bs.assignPlayerInput(playerID,'flyRelease',bs.WeakCall(self.onFlyRelease))

        self._connectedToPlayerID = playerID

    def curse(self):
        if not self._cursed:
            factory = getFactory()
            self._cursed = True
            # add the curse material..
            for attr in ['materials','rollerMaterials']:
                materials = getattr(self.node,attr)
                if not factory.curseMaterial in materials:
                    setattr(self.node,attr,materials + (factory.curseMaterial,))

            # -1 specifies no time limit
            if self.curseTime == -1:
                self.node.curseDeathTime = -1
            else:
                self.node.curseDeathTime = bs.getGameTime()+5000
                bs.gameTimer(5000,bs.WeakCall(self.curseExplode))

    def getBoxingGloves(self):
        self.node.boxingGloves = 1
        self.punchPowerScale = 1.7
        self.punchCooldown = 300

    def shieldsUp(self):
        factory = getFactory()
        if self.shield == None: 
            self.shield = bs.newNode('shield',owner=self.node,
                                     attrs={'color':(0.3,0.2,2.0),'radius':1.3})
            bs.connectAttr(self.node,'translateCenter',self.shield,'translate')
        self.shieldHitPoints = self.shieldHitPointsMax = 650
        self.shield.hurt = 0
        bs.playSound(factory.shieldUpSound,1.0,position=(self.node,'translate'))
        
    def handleMessage(self,m):

        if isinstance(m,bs.PickUpMessage):
            self.holdingNode = m.node

        elif isinstance(m,bs.DropMessage):
            self.holdingNode = None

        elif isinstance(m,bs.PickedUpMessage):
            bs.nodeMessage(self.node,"hurtSound")
            bs.nodeMessage(self.node,"pickedUp")
            # this counts as a hit
            self._numTimesHit += 1

        elif isinstance(m,bs.ShouldShatterMessage):
            # eww; seems we have to do this in a timer or it wont work right (since we're getting called from within update())
            bs.gameTimer(1,bs.WeakCall(self.shatter))

        elif isinstance(m,bs.ImpactDamageMessage):
            # eww; seems we have to do this in a timer or it wont work right (since we're getting called from within update())
            bs.gameTimer(1,bs.WeakCall(self._hitSelf,m.intensity))

        elif isinstance(m,bsPowerup.PowerupMessage):
            if self.dead: return True
            if self.pickUpPowerupCallback is not None:
                self.pickUpPowerupCallback(self)

            if (m.type == 'bomb'):
                tex = bsPowerup.getFactory().texBomb
                self._flashBillboard(tex)
                self._setBombCount(3)
                if self.powerupsExpire:
                    self.node.miniBillboard1Texture = tex
                    t = bs.getGameTime()
                    self.node.miniBillboard1StartTime = t
                    self.node.miniBillboard1EndTime = t+gPowerupWearOffTime
                    self._multiBombWearOffFlashTimer = bs.Timer(gPowerupWearOffTime-2000,bs.WeakCall(self._multiBombWearOffFlash))
                    self._multiBombWearOffTimer = bs.Timer(gPowerupWearOffTime,bs.WeakCall(self._multiBombWearOff))
            elif (m.type == 'speed'):
                self.node.runningShoes = 1
                self.setSpeed(max(self.speed,1.35))
            elif m.type == 'landMines':
                self.setLandMineCount(min(self.landMineCount+3,3))
            elif m.type == 'impactBombs':
                self.bombType = 'impact'
                tex = self._getBombTypeTex()
                self._flashBillboard(tex)
                if self.powerupsExpire:
                    self.node.miniBillboard2Texture = tex
                    t = bs.getGameTime()
                    self.node.miniBillboard2StartTime = t
                    self.node.miniBillboard2EndTime = t+gPowerupWearOffTime
                    self._bombWearOffFlashTimer = bs.Timer(gPowerupWearOffTime-2000,bs.WeakCall(self._bombWearOffFlash))
                    self._bombWearOffTimer = bs.Timer(gPowerupWearOffTime,bs.WeakCall(self._bombWearOff))
            elif m.type == 'stickyBombs':
                self.bombType = 'sticky'
                tex = self._getBombTypeTex()
                self._flashBillboard(tex)
                if self.powerupsExpire:
                    self.node.miniBillboard2Texture = tex
                    t = bs.getGameTime()
                    self.node.miniBillboard2StartTime = t
                    self.node.miniBillboard2EndTime = t+gPowerupWearOffTime
                    self._bombWearOffFlashTimer = bs.Timer(gPowerupWearOffTime-2000,bs.WeakCall(self._bombWearOffFlash))
                    self._bombWearOffTimer = bs.Timer(gPowerupWearOffTime,bs.WeakCall(self._bombWearOff))
            elif m.type == 'punch':
                self.hasBoxingGloves = True
                tex = bsPowerup.getFactory().texPunch
                self._flashBillboard(tex)
                self.getBoxingGloves()
                if self.powerupsExpire:
                    self.node.boxingGlovesFlashing = 0
                    self.node.miniBillboard3Texture = tex
                    t = bs.getGameTime()
                    self.node.miniBillboard3StartTime = t
                    self.node.miniBillboard3EndTime = t+gPowerupWearOffTime
                    self._boxingGlovesWearOffFlashTimer = bs.Timer(gPowerupWearOffTime-2000,bs.WeakCall(self._glovesWearOffFlash))
                    self._boxingGlovesWearOffTimer = bs.Timer(gPowerupWearOffTime,bs.WeakCall(self._glovesWearOff))
            elif m.type == 'shield':
                self.shieldsUp()
            elif m.type == 'curse':
                self.curse()
            elif (m.type == 'iceBombs'):
                self.bombType = 'ice'
                tex = self._getBombTypeTex()
                self._flashBillboard(tex)
                if self.powerupsExpire:
                    self.node.miniBillboard2Texture = tex
                    t = bs.getGameTime()
                    self.node.miniBillboard2StartTime = t
                    self.node.miniBillboard2EndTime = t+gPowerupWearOffTime
                    self._bombWearOffFlashTimer = bs.Timer(gPowerupWearOffTime-2000,bs.WeakCall(self._bombWearOffFlash))
                    self._bombWearOffTimer = bs.Timer(gPowerupWearOffTime,bs.WeakCall(self._bombWearOff))
            elif (m.type == 'health'):
                if self._cursed:
                    self._cursed = False
                    # remove cursed material
                    factory = getFactory()
                    for attr in ['materials','rollerMaterials']:
                        materials = getattr(self.node,attr)
                        if factory.curseMaterial in materials:
                            setattr(self.node,attr,tuple(m for m in materials if m != factory.curseMaterial))
                    
                    self.node.curseDeathTime = 0
                self.hitPoints = self.hitPointsMax
                self._flashBillboard(bsPowerup.getFactory().texHealth)
                self.node.hurt = 0
                self._lastHitTime = None
                self._numTimesHit = 0
                
            bs.nodeMessage(self.node,"flash")
            if m.source is not None: bs.nodeMessage(m.source,bsPowerup.AcceptMessage())
            return True

        elif isinstance(m,bs.FreezeMessage):
            if self.node.invincible == True:
                bs.playSound(getFactory().blockSound,1.0,position=(self.node,'translate'))
                return
            if self.shield != None: return
            if not self.frozen:
                self.frozen = True
                self.node.frozen = 1
                bs.gameTimer(5000,bs.WeakCall(bs.nodeMessage,self.node,bs.ThawMessage()))
                # instantly shatter if we're already dead (otherwise its hard to tell we're dead)
                if self.hitPoints <= 0:
                    self.shatter()


        elif isinstance(m,bs.ThawMessage):
            if self.frozen and not self.shattered:
                self.frozen = False
                self.node.frozen = 0
                
        elif isinstance(m,bs.HitMessage):
            if self.node.invincible == True:
                bs.playSound(getFactory().blockSound,1.0,position=(self.node,'translate'))
                return True

            # if we were recently hit, don't count this as another
            # (so punch flurries and bomb pileups essentially count as 1 hit)
            gameTime = bs.getGameTime()
            if self._lastHitTime == None or gameTime-self._lastHitTime > 1000:
                self._numTimesHit += 1
                self._lastHitTime = gameTime
            
            mag = m.magnitude
            velocityMag = m.velocityMagnitude

            damageScale = 0.22

            # if they've got a shield, deliver it to that instead..
            if self.shield != None:

                if m.flatDamage: damage = m.flatDamage
                else:
                    # hit our spaz with an impulse but tell it to only return theoretical damage; not apply the impulse..
                    bs.nodeMessage(self.node,"impulse",m.pos[0],m.pos[1],m.pos[2],
                                   m.velocity[0],m.velocity[1],m.velocity[2],
                                   mag,velocityMag,m.radius,1,m.forceDirection[0],m.forceDirection[1],m.forceDirection[2])
                    damage = damageScale * self.node.damage

                self.shieldHitPoints -= damage

                self.shield.hurt = 1.0 - self.shieldHitPoints/self.shieldHitPointsMax
                # its a cleaner event if a hit just kills the shield without damaging the player..
                # however, massive damage events should still be able to damage the player..
                # this hopefully gives us a happy medium.
                maxSpillover = 500
                if self.shieldHitPoints <= 0:
                    # fixme - transition out perhaps?..
                    bs.deleteNode(self.shield)
                    self.shield = None
                    bs.playSound(getFactory().shieldDownSound,1.0,position=(self.node,'translate'))
                    # emit some cool lookin sparks when the shield dies
                    t = self.node.translate
                    bs.emitBGDynamics(position=(t[0],t[1]+0.9,t[2]),
                                      velocity=self.node.velocity,
                                      count=random.randrange(20,30),scale=1.0,spread=0.6,chunkType='spark')

                else:
                    bs.playSound(getFactory().shieldHitSound,0.5,position=(self.node,'translate'))

                # emit some cool lookin sparks on shield hit
                bs.emitBGDynamics(position=m.pos,
                                  velocity=(m.forceDirection[0]*1.0,
                                            m.forceDirection[1]*1.0,
                                            m.forceDirection[2]*1.0),
                                  count=min(30,5+int(damage*0.005)),scale=0.5,spread=0.3,chunkType='spark')


                # if they passed our spillover threshold, pass damage along to spaz
                if self.shieldHitPoints <= -maxSpillover:
                    leftoverDamage = -maxSpillover-self.shieldHitPoints
                    shieldLeftoverRatio = leftoverDamage/damage

                    # scale down the magnitudes applied to spaz accordingly..
                    mag *= shieldLeftoverRatio
                    velocityMag *= shieldLeftoverRatio
                else:
                    return True # good job shield!
            else: shieldLeftoverRatio = 1.0

            if m.flatDamage:
                damage = m.flatDamage * shieldLeftoverRatio
            else:
                # hit it with an impulse and get the resulting damage
                bs.nodeMessage(self.node,"impulse",m.pos[0],m.pos[1],m.pos[2],
                               m.velocity[0],m.velocity[1],m.velocity[2],
                               mag,velocityMag,m.radius,0,m.forceDirection[0],m.forceDirection[1],m.forceDirection[2])

                damage = damageScale * self.node.damage
            bs.nodeMessage(self.node,"hurtSound")

            # play punch impact sound based on damage if it was a punch
            if m.hitType == 'punch':

                self.onPunched(damage)

                # if damage was significant, lets show it
                if damage > 350: bsUtils.showDamageCount(self.getActivity(),'-'+str(int(damage/10))+"%",m.pos,m.forceDirection)
                                               
                # lets always add in a super-punch sound with boxing gloves just to differentiate them
                if m.hitSubType == 'superPunch':
                    bs.playSound(getFactory().superPunchSound,1.0,
                                 position=(self.node,'translate'))

                if damage > 500: sound = getFactory().punchSoundStrong
                #elif damage > 300: sound = self.factory.punchSound
                #else: sound = self.factory.punchSoundWeak
                else: sound = getFactory().punchSound
                bs.playSound(sound,1.0,position=(self.node,'translate'))

                # throw up some chunks
                bs.emitBGDynamics(position=m.pos,
                                  velocity=(m.forceDirection[0]*0.5,
                                            m.forceDirection[1]*0.5,
                                            m.forceDirection[2]*0.5),
                                  count=min(10,1+int(damage*0.0025)),scale=0.3,spread=0.03);

                bs.emitBGDynamics(position=m.pos,
                                  chunkType='sweat',
                                  velocity=(m.forceDirection[0]*1.3,
                                            m.forceDirection[1]*1.3+5.0,
                                            m.forceDirection[2]*1.3),
                                  count=min(30,1+int(damage*0.04)),
                                  scale=0.9,
                                  spread=0.28);
                # momentary flash
                hurtiness = damage*0.003
                punchPos = (m.pos[0]+m.forceDirection[0]*0.02,
                            m.pos[1]+m.forceDirection[1]*0.02,
                            m.pos[2]+m.forceDirection[2]*0.02)
                flashColor = (1.0,0.8,0.4)
                light = bs.newNode("light",
                                   owner=self.getActivity(),
                                   attrs={'translate':punchPos,
                                          'radius':0.12+hurtiness*0.12,
                                          'intensity':0.3*(1.0+1.0*hurtiness),
                                          'heightAttenuated':False,
                                          'color':flashColor})
                bs.gameTimer(60,bs.WeakCall(bs.deleteNode,light))


                flash = bs.newNode("flash",
                                   owner=self.getActivity(),
                                   attrs={'translate':punchPos,
                                          'size':0.17+0.17*hurtiness,
                                          'color':flashColor})
                bs.gameTimer(60,bs.WeakCall(bs.deleteNode,flash))

            if m.hitType == 'impact':
                bs.emitBGDynamics(position=m.pos,
                                  velocity=(m.forceDirection[0]*2.0,
                                            m.forceDirection[1]*2.0,
                                            m.forceDirection[2]*2.0),
                                  count=min(10,1+int(damage*0.01)),scale=0.4,spread=0.1);
                
            if self.hitPoints > 0:

                # its kinda crappy to die from impacts, so lets reduce impact damage
                # by a reasonable amount if it'll keep us alive
                if m.hitType == 'impact' and damage > self.hitPoints:
                    # drop damage to whatever puts us at 10 hit points, or 200 less than it used to be
                    # whichever is greater (so it *can* still kill us if its high enough)
                    newDamage = max(damage-200,self.hitPoints-10)
                    damage = newDamage

                bs.nodeMessage(self.node,"flash")
                # if we're holding something, drop it
                if damage > 0.0 and self.holdingNode != None:
                    bs.nodeMessage(self.node,"drop")
                self.hitPoints -= damage
                self.node.hurt = 1.0 - self.hitPoints/self.hitPointsMax
                # if we're cursed, *any* damage blows us up
                if self._cursed and damage > 0:
                    bs.gameTimer(50,bs.WeakCall(self.curseExplode,m.source))
                # if we're frozen, shatter.. otherwise die if we hit zero
                if self.frozen and (damage > 200 or self.hitPoints <= 0):
                    self.shatter()
                elif self.hitPoints <= 0:
                    bs.nodeMessage(self.node,bs.DieMessage(how='impact'))

            # if we're dead, take a look at the smoothed damage val
            # (which gives us a smoothed average of recent damage) and shatter
            # us if its grown high enough
            if self.hitPoints <= 0:
                damageAvg = self.node.damageSmoothed * damageScale
                if damageAvg > 1000:
                    self.shatter()

        elif isinstance(m,BombDeathMsg):
            self.bombCount += 1
        
        elif isinstance(m,AttackMessage):

            if self.frozen or self.node.knockout > 0.0: return True

            if m.type == 'bomb':
                if self.holdingNode == None:
                    self.dropBomb()
                else:
                    bs.nodeMessage(self.node,'pickUp')
                    
            elif m.type == 'punch':
                if self.punchCallback is not None:
                    self.punchCallback(self)

                t = bs.getGameTime()
                self._punchedNodes = set() # reset this..
                if t - self.lastPunchTime > self.punchCooldown:
                    self.lastPunchTime = t
                    bs.nodeMessage(self.node,'punch')
                    if self.holdingNode == None:
                        bs.gameTimer(100,bs.WeakCall(bs.playSound,getFactory().swishSound,0.3,(self.node,'translate')))

        elif isinstance(m,bs.DieMessage):
            wasDead = self.dead
            self.dead = True
            self.hitPoints = 0
            if m.immediate:
                bs.deleteNode(self.node)
            elif self.node.exists():
                self.node.hurt = 1.0
                if self.playBigDeathSound and not wasDead:
                    bs.playSound(getFactory().singlePlayerDeathSound)
                self.node.dead = True
                bs.gameTimer(2000,bs.WeakCall(bs.deleteNode,self.node))

        elif isinstance(m,bs.OutOfBoundsMessage):
            # by default we just die here
            self.handleMessage(bs.DieMessage(how='fall'))

        elif isinstance(m,bs.StandMessage):
            self._lastStandPos = (m.pos[0],m.pos[1],m.pos[2])
            bs.nodeMessage(self.node,"stand",m.pos[0],m.pos[1],m.pos[2],m.angle)

        elif isinstance(m,CurseExplodeMessage):
            self.curseExplode()

        elif isinstance(m,PunchHitMessage):

            node = bs.getCollideValues("opposingNode")

            # only allow one hit per node per punch
            if node is not None and node.exists() and not node in self._punchedNodes:

                punchMomentumAngular = self.node.punchMomentumAngular * self.punchPowerScale
                punchPower = self.node.punchPower * self.punchPowerScale

                # ok here's the deal:  we pass along our base velocity for use in the
                # impulse damage calculations since that is a more predictable value
                # than our fist velocity, which is rather erratic.
                # ...however we want to actually apply force in the direction our fist
                # is moving so it looks better.. so we still pass that along as a direction
                # ..perhaps a time-averased fist-velocity would work too?.. should try that.
                
                # if its something besides another spaz, just do a muffled punch sound
                if node.getType() != 'spaz':
                    bs.playSound(getFactory().medImpactSounds,1.0,position=(self.node,'translate'))

                t = self.node.punchTranslate
                punchDir = self.node.punchVelocity
                v = self.node.punchMomentumLinear

                self._punchedNodes.add(node)
                bs.nodeMessage(node,bs.HitMessage(pos=t,
                                                  velocity=v,
                                                  magnitude=punchPower*punchMomentumAngular*110.0,
                                                  velocityMagnitude=punchPower*40,
                                                  radius=0,
                                                  srcNode=self.node,
                                                  source=self.source,
                                                  forceDirection = punchDir,
                                                  hitType='punch',
                                                  hitSubType='superPunch' if self.hasBoxingGloves else 'default'))

                # also apply opposite to ourself for the first punch only
                # ..this is given as a constant force so that it is more noticable for slower punches
                # where it matters.. for fast awesome looking punches its ok if we punch 'through' the target
                mag = -400.0
                if self.hockey: mag *= 0.5
                if len(self._punchedNodes) == 1:  bs.nodeMessage(self.node,"kickBack",t[0],t[1],t[2],
                                                                 punchDir[0],punchDir[1],punchDir[2],mag)

        elif isinstance(m,PickupMessage):
            opposingNode,opposingBody = bs.getCollideValues("opposingNode","opposingBody")

            if not opposingNode.exists(): return True

            # dont allow picking up of invincible dudes
            try:
                if opposingNode.invincible == True: return True
            except Exception: pass

            # if we're grabbing the pelvis of a non-shattered spaz, we wanna grab the torso instead
            if opposingNode.getType() == 'spaz' and not opposingNode.shattered and opposingBody == 4:
                opposingBody = 1

            # special case.. wanna make sure we don't pick up bombs that we wanted to drop
            if self._dontPickThisUp == opposingNode: return True

            # special case - if we're holding a flag, dont replace it
            # ( hmm - should make this customizable or more low level )
            held = self.node.holdNode

            if held is not None and held.exists() and held.getType() == "flag":
                return True

            self.node.holdNode = opposingNode
            self.node.holdBody = opposingBody
        else:
            bs.Actor.handleMessage(self,m)

    def dropBomb(self):

        if self.dropBombCallback is not None: self.dropBombCallback(self)

        if (self.landMineCount <= 0 and self.bombCount <= 0) or self.frozen: return
        bs.playSound(getFactory().popSound,volume=0.2,position=(self.node,"translate"))
        p = self.node.translateForward
        v = self.node.velocity

        if self.landMineCount > 0:
            droppingBomb = False
            self.setLandMineCount(self.landMineCount-1)
            bombType = 'landMine'
        else:
            droppingBomb = True
            bombType = self.bombType

        b = bsBomb.Bomb(self._activity(),bs.Actor.RETAINED,
                        translate=(p[0],p[1] - 0.0,p[2]),
                        velocity=(v[0],v[1],v[2]),
                        bombType=bombType,
                        blastRadius=self.blastRadius,
                        source=self.source,
                        owner=self.node)

        if droppingBomb:
            self.bombCount -= 1
            bs.addNodeDeathAction(b.node,bs.WeakCall(bs.nodeMessage,self.node,BombDeathMsg()))

        bs.gameTimer(1,bs.WeakCall(self._pickUp,b.node))
        self._dontPickThisUp = None

    def _pickUp(self,node):
        if self.node.exists() and node.exists():
            self.node.holdNode = node
            self.node.holdBody = 0
        
    def setLandMineCount(self,count):
        self.landMineCount = count
        if self.node.exists():
            if self.landMineCount != 0:
                self.node.counterText = 'x'+str(self.landMineCount)
                self.node.counterTexture = bsPowerup.getFactory().texLandMines
            else:
                self.node.counterText = ''

    def curseExplode(self,source=None):
        if self._cursed and self.node.exists():
            self.shatter(extreme=True)
            bs.nodeMessage(self.node,bs.DieMessage())
            activity = self._activity()
            if activity:
                bsBomb.Blast(activity,bs.Actor.RETAINED,
                             translate=self.node.translate,
                             velocity=self.node.velocity,
                             blastRadius=3.0,type='normal',source=self.source if source in [None,-1] else source)
            self._cursed = False

    def shatter(self,extreme=False):
        if self.shattered: return
        self.shattered = True
        if self.frozen:
            # momentary flash of light
            light = bs.newNode("light",
                               owner=self.getActivity(),
                               attrs={"translate":self.node.translate,"radius":0.5,"heightAttenuated":False,"color": (0.8,0.8,1.0)})
            bsUtils.animate(light,"intensity",{0:3.0, 40:0.5, 80:0.07, 300:0})
            bs.gameTimer(300,bs.WeakCall(bs.deleteNode,light))
            # emit ice chunks..
            bs.emitBGDynamics(position=self.node.translate,
                              velocity=self.node.velocity,
                              count=int(random.random()*10.0+10.0),scale=0.6,spread=0.2,chunkType='ice');
            bs.emitBGDynamics(position=self.node.translate,
                              velocity=self.node.velocity,
                              count=int(random.random()*10.0+10.0),scale=0.3,spread=0.2,chunkType='ice');

            bs.playSound(getFactory().shatterSound,1.0,position=(self.node,'translate'))
        else:
            bs.playSound(getFactory().splatterSound,1.0,position=(self.node,'translate'))
        bs.nodeMessage(self.node,bs.DieMessage())
        self.node.shattered = 2 if extreme else 1

    def _hitSelf(self,intensity):

        # clean exit if we're dead..
        try: pos = self.node.translate
        except Exception: return

        bs.nodeMessage(self.node,bs.HitMessage(flatDamage=50.0*intensity,
                                               pos=pos,
                                               forceDirection=self.node.velocity,
                                               hitType='impact'))
        bs.nodeMessage(self.node,"knockout",max(0.0,50.0*intensity))
        if intensity > 5: s = getFactory().bigImpactSounds
        elif intensity > 3: s = getFactory().hardImpactSounds
        else: s = getFactory().medImpactSounds
        bs.playSound(s,position=pos,volume=5.0)
        
    def _getBombTypeTex(self):
        bombFactory = bsPowerup.getFactory()
        if self.bombType == 'sticky': return bombFactory.texStickyBombs
        elif self.bombType == 'ice': return bombFactory.texIceBombs
        elif self.bombType == 'impact': return bombFactory.texImpactBombs
        else: raise Exception()
        
    def _flashBillboard(self,tex):
        self.node.billboardTexture = tex
        self.node.billboardFlashing = 0
        bsUtils.animate(self.node,"billboardOpacity",{0:0.0,100:1.0,400:1.0,500:0.0})

    def _setBombCount(self,count):
        # we cant just set bombCount cuz some bombs may be laid currently
        # so we have to do a relative diff based on max
        diff = count - self.maxBombCount
        self.maxBombCount += diff
        self.bombCount += diff

    def _glovesWearOffFlash(self):
        if self.node.exists():
            self.node.boxingGlovesFlashing = 1
            self.node.billboardTexture = bsPowerup.getFactory().texPunch
            self.node.billboardOpacity = 1
            self.node.billboardFlashing = 1

    def _glovesWearOff(self):
        self.punchPowerScale = gBasePunchPowerScale
        self.punchCooldown = gBasePunchCooldown
        self.hasBoxingGloves = False
        if self.node.exists():
            bs.playSound(bsPowerup.getFactory().powerdownSound,position=(self.node,'translate'))
            self.node.boxingGloves = 0
            self.node.billboardOpacity = 0.0

    def _multiBombWearOffFlash(self):
        if self.node.exists():
            self.node.billboardTexture = bsPowerup.getFactory().texBomb
            self.node.billboardOpacity = 1
            self.node.billboardFlashing = 1

    def _multiBombWearOff(self):
        self._setBombCount(1)
        if self.node.exists():
            bs.playSound(bsPowerup.getFactory().powerdownSound,position=(self.node,'translate'))
            self.node.billboardOpacity = 0.0

    def _bombWearOffFlash(self):
        if self.node.exists():
            self.node.billboardTexture = self._getBombTypeTex()
            self.node.billboardOpacity = 1
            self.node.billboardFlashing = 1

    def _bombWearOff(self):
        self.bombType = self.bombTypeDefault
        if self.node.exists():
            bs.playSound(bsPowerup.getFactory().powerdownSound,position=(self.node,'translate'))
            self.node.billboardOpacity = 0.0



class PlayerSpaz(Spaz):
    """spaz subclass with some high-level game-event hooks added"""
    def __init__(self,activity,retain,color=(1,1,1),highlight=(0.5,0.5,0.5),character="Spaz",player=None,powerupsExpire=False,hockey=False):
        Spaz.__init__(self,activity,retain,color=color,highlight=highlight,character=character,source=player.playerID,fly=bs.globals.happyThoughts,startInvincible=True,powerupsExpire=powerupsExpire,hockey=hockey)
        self.lastAttackerID = None
        self.lastAttackedTime = 0
        self.lastAttackedType = None
        self.heldCount = 0
        self.lastPlayerHeldByID = None
        self._player = weakref.ref(player)

    def __superHandleMessage(self,m):
        super(PlayerSpaz,self).handleMessage(m)
        
    def getPlayer(self):
        return self._player()

    def updateForPlayer(self):
        pass

    def handleMessage(self,m):

        # keep track of if we're being held and by who most recently
        if isinstance(m,bs.PickedUpMessage):
            self.__superHandleMessage(m) # augment standard behavior
            self.heldCount += 1
            pickedUpBy = m.node.sourcePlayer
            if pickedUpBy != -1:
                self.lastPlayerHeldByID = pickedUpBy

        elif isinstance(m,bs.DroppedMessage):
            self.__superHandleMessage(m) # augment standard behavior
            self.heldCount -= 1
            if self.heldCount < 0:
                print "ERROR: spaz heldCount < 0"
            # let's count someone dropping us as an attack..
            try: pickedUpBy = m.node.sourcePlayer
            except Exception: pickedUpBy = -1
            if pickedUpBy != -1:
                self.lastAttackerID = pickedUpBy
                self.lastAttackedTime = bs.getGameTime()
                self.lastAttackedType = ('pickedUp','default')
            
        elif isinstance(m,bs.DieMessage):

            # report player deaths to the game
            if not self.dead:

                # immediate or left-game kills don't count as 'kills'
                killed = (m.immediate==False and m.how!='leftGame')

                activity = self._activity()

                if not killed:
                    killerPlayer = None
                else:
                    # if this player was being held at the time of death, the holder is the killer
                    if self.heldCount > 0 and self.lastPlayerHeldByID is not None:
                        try: killerPlayer = activity.players[self.lastPlayerHeldByID]
                        except Exception: killerPlayer = None
                    else:
                        # otherwise if they were attacked by someone in the last few seconds
                        # that person's the killer.. otherwise it was a suicide
                        if self.lastAttackerID is not None and bs.getGameTime() - self.lastAttackedTime < 4000:
                            try: killerPlayer = activity.players[self.lastAttackerID]
                            except Exception: killerPlayer = None
                        else:
                            killerPlayer = self.getPlayer()
                if activity:
                    activity.handleMessage(PlayerSpazDeathMessage(self,killed,killerPlayer,m.how))
            self.__superHandleMessage(m) # augment standard behavior

        # keep track of the player who last hit us for point rewarding
        elif isinstance(m,bs.HitMessage):
            if m.source is not None:
                self.lastAttackerID = m.source
                self.lastAttackedTime = bs.getGameTime()
                self.lastAttackedType = (m.hitType,m.hitSubType)
            self.__superHandleMessage(m) # augment standard behavior
            activity = self._activity()
            if activity: activity.handleMessage(PlayerSpazHurtMessage(self))
        else:
            Spaz.handleMessage(self,m)


class RespawnIcon(object):

    def __init__(self,player,activity,respawnTime):

        onRight = False
        self._visible = True
        sessionType = activity.getSession().getSessionType()
        if sessionType == 'teams':
            onRight = player.getTeam().teamID%2==1

            # store a list of icons in the team
            try: respawnIcons = player.getTeam().data['_spazRespawnIconsRight']
            except Exception: respawnIcons = player.getTeam().data['_spazRespawnIconsRight'] = {}
            offsExtra = 0
        else:
            onRight = False

            # store a list of icons in the activity
            try: respawnIcons = activity._spazRespawnIconsRight
            except Exception: respawnIcons = activity._spazRespawnIconsRight = {}

            if sessionType == 'freeForAll': offsExtra = -110
            else: offsExtra = 0

        # now find the first unused slot and use that
        index = 0
        while index in respawnIcons and respawnIcons[index]() is not None and respawnIcons[index]()._visible: index += 1
        respawnIcons[index] = weakref.ref(self)

        offs = offsExtra + index*-45
        icon = player.getIcon()
        texture = icon['texture']

        self._image = bs.NodeActor(activity,bs.Actor.UNRETAINED,
                                   'image',{'texture':texture,
                                            'translate':(-40 if onRight else 40,-180+offs),
                                            'scale':(32,32),
                                            'opacity':0.7,
                                            'color':(1,0.4,0.4),
                                            'absoluteScale':True,
                                            'attach':'topRight' if onRight else 'topLeft'})
        bsUtils.animate(self._image.node,'opacity',{0:0,200:0.7})

        self._name = bs.NodeActor(activity,bs.Actor.UNRETAINED,
                                  'text',{'vAttach':'top',
                                          'hAttach':'right' if onRight else 'left',
                                          'text':player.getName(),
                                          'maxWidth':120,
                                          'hAlign':'center',
                                          'vAlign':'center',
                                          'color':(1.0,0.3,0.3,1.0),
                                          'scale':[0.5],
                                          'translate':(-40 if onRight else 40,-205+42+offs)})
        bsUtils.animate(self._name.node,'scale',{0:0,100:0.5})

        self._text = bs.NodeActor(activity,bs.Actor.UNRETAINED,
                                  'text',{'translate':(-60 if onRight else 60,-192+offs),
                                          'hAttach':'right' if onRight else 'left',
                                          'hAlign':'right' if onRight else 'left',
                                          'scale':[0.9],
                                          'vAttach':'top',
                                          'color':(1.0,0.3,0.3,1),
                                          'text':''})
        bsUtils.animate(self._text.node,'scale',{0:0,100:0.9})

        self._respawnTime = bs.getGameTime()+respawnTime
        self._update()
        self._timer = bs.Timer(1000,bs.WeakCall(self._update),repeat=True)

    def _update(self):
        remaining = int(round(self._respawnTime-bs.getGameTime())/1000.0)
        if remaining > 0:
            if self._text.node.exists():
                self._text.node.text = str(remaining)
        else: self._clear()
            
    def _clear(self):
        self._visible = False
        self._image = self._text = self._timer = self._name = None
        
class SpazBotDeathMessage(object):
    """A bad guy has died."""
    def __init__(self,badGuy,killer,how):
        self.badGuy = badGuy
        self.killer = killer
        self.how = how

class SpazBotPunchedMessage(object):
    """A bad guy got punched."""
    def __init__(self,badGuy,damage):
        self.badGuy = badGuy
        self.damage = damage

class SpazBot(Spaz):
    """ a really dumb AI version of spaz """

    character = 'Spaz'
    punchiness = 0.5
    throwiness = 0.7
    static = False
    run = False
    chargeDistMin = 0.0 # when we can start a new charge
    chargeDistMax = 2.0 # when we can start a new charge
    runDistMin = 0.0 # how close we can be to continue running
    chargeSpeedMin = 0.4
    chargeSpeedMax = 1.0
    throwDistMin = 5.0
    throwDistMax = 9.0
    throwRate = 1.0
    defaultBombType = 'normal'
    defaultBombCount = 3
    startCursed = False
    color=gDefaultBotColor
    highlight=gDefaultBotHighlight

    def __init__(self,activity,retain,level):
        Spaz.__init__(self,activity,retain,color=self.color,highlight=self.highlight,character=self.character,
                                   source=-1,fly=False,startInvincible=False,canAcceptPowerups=False)

        # if you need to add custom behavior to a bot, set this to a callable which takes one
        # arg (the bot) and returns False if the bot's normal update should be run and True if not
        self.updateCallback = None
        self._level = weakref.ref(level)

        self.lastAttackerID = None
        self.lastAttackedTime = 0
        self.lastAttackedType = None
        self.targetPointDefault = None
        self.heldCount = 0
        self.lastPlayerHeldByID = None
        self.targetFlag = None
        self._chargeSpeed = 0.5*(self.chargeSpeedMin+self.chargeSpeedMax)
        self._leadAmount = 0.5
        self._mode = 'wait'
        self._chargeClosingIn = False
        self._lastChargeDist = 0.0
        self._running = False
        self._lastJumpTime = 0

        if self.startCursed: self.curse()
            
    def _getTargetPlayerPt(self):
        """ returns the default player pt we're targeting """
        bp = bsVector.Vector(*self.node.translate)
        closestLen = None
        closestVel = None
        for pp,pv in self._playerPts:

            l = (pp-bp).length()
            # ignore player-points that are significantly below the bot
            # (keeps bots from following players off cliffs)
            if (closestLen == None or l < closestLen) and (pp[1] > bp[1] - 5.0):
                closestLen = l
                closestVel = pv
                closest = pp
        if closestLen != None:
            return (bsVector.Vector(closest[0],closest[1],closest[2]),
                    bsVector.Vector(closestVel[0],closestVel[1],closestVel[2]))
        else:
            return None,None

    def setPlayerPts(self,pts):
        self._playerPts = pts

    def updateAI(self):

        if self.updateCallback != None:
            if self.updateCallback(self) == True: return # true means bot has been handled

        t = self.node.translate
        ourPos = bsVector.Vector(t[0],0,t[2])
        canAttack = True

        # if we're a flag-bearer, we're pretty simple-minded - just walk towards the flag and try to pick it up
        if self.targetFlag is not None:

            if not self.targetFlag.node.exists():
                # our flag musta died :-C
                self.targetFlag = None
                return
            if self.holdingNode is not None:
                try: holdingFlag = self.holdingNode.getType() == 'flag'
                except Exception: holdingFlag = False
            else: holdingFlag = False
            # if we're holding the flag, just walk left
            if holdingFlag:
                # just walk left
                bs.nodeMessage(self.node,'moveLeftRight',-1.0 * 32000)
                bs.nodeMessage(self.node,'moveUpDown',0)
            # otherwise try to go pick it up
            else:
                targetPtRaw = bsVector.Vector(*self.targetFlag.node.translate)
                targetVel = bsVector.Vector(0,0,0)
                diff = (targetPtRaw-ourPos)
                diff = bsVector.Vector(diff[0],0,diff[2]) # dont care about y
                dist = diff.length()
                toTarget = diff.normal()

                # if we're holding some non-flag item, drop it
                if self.holdingNode != None:
                    bs.nodeMessage(self.node,'pickUp')
                    return

                # if we're a runner, run only when not super-near the flag
                if self.run and dist > 3.0:
                    self._running = True
                    bs.nodeMessage(self.node,'run')
                else:
                    self._running = False
                    bs.nodeMessage(self.node,'runRelease')

                bs.nodeMessage(self.node,'moveLeftRight',toTarget.x() * 32000)
                bs.nodeMessage(self.node,'moveUpDown',toTarget.z() * -32000)
                # if we're close to it, pick it up
                if dist < 1.25: bs.nodeMessage(self.node,'pickUp')
            return
        # not a flag-bearer.. if we're holding anything but a bomb, drop it
        else:
            if self.holdingNode != None:
                try: holdingBomb = self.holdingNode.getType() in ['bomb','prop']
                except Exception: holdingBomb = False
                if not holdingBomb:
                    bs.nodeMessage(self.node,'pickUp')
                    return
            #else: holdingBomb = False
            
        targetPtRaw,targetVel = self._getTargetPlayerPt()

        if targetPtRaw is None:
            # use default target if we've got one
            if self.targetPointDefault is not None:
                targetPtRaw = self.targetPointDefault
                targetVel = bsVector.Vector(0,0,0)
                canAttack = False
            # with no target, we stop moving and drop whatever we're holding
            else:
                bs.nodeMessage(self.node,'moveLeftRight',0)
                bs.nodeMessage(self.node,'moveUpDown',0)
                if self.holdingNode != None:
                    bs.gameTimer(1,bs.WeakCall(bs.nodeMessage,self.node,'pickUp')) # throws
                return

        # we dont want height to come into play
        targetPtRaw.data[1] = 0
        targetVel.data[1] = 0

        distRaw = (targetPtRaw-ourPos).length()
        # use a point out in front of them as real target (more out in front the farther from us they are)
        targetPt = targetPtRaw + targetVel*distRaw*0.3*self._leadAmount

        diff = (targetPt-ourPos)
        dist = diff.length()
        toTarget = diff.normal()

        if self._mode == 'throw':
            # we can only throw if alive and well..
            if not self.dead and not self.node.knockout:

                timeTillThrow = self._throwReleaseTime-bs.getGameTime()
                if self.holdingNode == None:
                    # if we havnt thrown yet, whip out the bomb
                    if not self._haveDroppedThrowBomb:
                        self.dropBomb()
                        self._haveDroppedThrowBomb = True
                    # otherwise our lack of held node means we successfully released our bomb.. lets retreat now
                    else:
                        self._mode = 'flee'

                # oh crap we're holding a bomb.. better throw it.
                elif timeTillThrow <= 0:
                    # jump and throw..
                    if dist > 5.0:
                        bs.nodeMessage(self.node,'jump')
                        bs.gameTimer(100,bs.WeakCall(bs.nodeMessage,self.node,'pickUp')) # throws
                    else:
                        bs.gameTimer(1,bs.WeakCall(bs.nodeMessage,self.node,'pickUp')) # throws

                if self.static:
                    if timeTillThrow < 300: speed = 1.0
                    elif timeTillThrow < 700 and dist > 3.0: speed = -1.0 # whiplash for long throws
                    else: speed = 0.02
                else:
                    if timeTillThrow < 700:
                        # right before throw charge full speed towards target
                        speed = 1.0
                    else:
                        # earlier we can hold or move backward for a whiplash
                        speed = 0.0125
                bs.nodeMessage(self.node,'moveLeftRight',toTarget.x() * 32000*speed)
                bs.nodeMessage(self.node,'moveUpDown',toTarget.z() * -32000*speed)

        elif self._mode == 'charge':
            if random.random() < 0.3:
                self._chargeSpeed = random.uniform(self.chargeSpeedMin,self.chargeSpeedMax)

                # if we're a runner we run during charges *except when near an edge (otherwise we tend to fly off easily)
                if self.run and distRaw > self.runDistMin:
                    self._leadAmount = 0.3
                    self._running = True
                    bs.nodeMessage(self.node,'run')
                else:
                    self._leadAmont = 0.01
                    self._running = False
                    bs.nodeMessage(self.node,'runRelease')

            bs.nodeMessage(self.node,'moveLeftRight',toTarget.x() * 32000*self._chargeSpeed)
            bs.nodeMessage(self.node,'moveUpDown',toTarget.z() * -32000*self._chargeSpeed)

        elif self._mode == 'wait':
            # every now and then, aim towards our target.. other than that, just stand there
            if bs.getGameTime()%1234 < 100:
                bs.nodeMessage(self.node,'moveLeftRight',toTarget.x() * 400)
                bs.nodeMessage(self.node,'moveUpDown',toTarget.z() * -400)
            else:
                bs.nodeMessage(self.node,'moveLeftRight',0)
                bs.nodeMessage(self.node,'moveUpDown',0)

        elif self._mode == 'flee':

            # even if we're a runner, only run till we get away from our target (if we keep running we tend to run off edges)
            if self.run and dist < 3.0:
                self._running = True
                bs.nodeMessage(self.node,'run')
            else:
                self._running = False
                bs.nodeMessage(self.node,'runRelease')
            bs.nodeMessage(self.node,'moveLeftRight',toTarget.x() * -32000)
            bs.nodeMessage(self.node,'moveUpDown',toTarget.z() * 32000)

        # we might wanna switch states unless we're doing a throw (in which case thats our sole concern)
        if self._mode != 'throw':

            # if we're currently charging, keep track of how far we are from our target..
            # when this value increases it means our charge is over (ran by them or something)
            if self._mode == 'charge':
                if self._chargeClosingIn and dist < 3.0 and dist > self._lastChargeDist:
                    self._chargeClosingIn = False
                self._lastChargeDist = dist

            # if we have a clean shot, throw!
            if dist >= self.throwDistMin and dist < self.throwDistMax and random.random() < self.throwiness and canAttack:
                self._mode = 'throw'
                self._leadAmount = (0.4+random.random()*0.6) if distRaw > 4.0 else (0.1+random.random()*0.4)
                self._haveDroppedThrowBomb = False
                self._throwReleaseTime = bs.getGameTime() + (1.0/self.throwRate)*(800 + int(1300*random.random()))

            # if we're static, always charge (which for us means barely move)
            elif self.static:
                self._mode = 'wait'
                
            # if we're too close to charge (and arent in the middle of an existing charge) run away
            elif dist < self.chargeDistMin and not self._chargeClosingIn:
                # ..unless we're near an edge, in which case we got no choice but to charge..
                #if self.game.isPointNearEdge(ourPos.x(),ourPos.z(),self._running):
                if self._level().isPointNearEdge(ourPos,self._running):
                    if self._mode != 'charge':
                        self._mode = 'charge'
                        self._leadAmount = 0.2
                        self._chargeClosingIn = True
                        self._lastChargeDist = dist
                else:
                    self._mode = 'flee'

            # we're within charging distance, backed against an edge, or farther than our max throw distance.. chaaarge!
            elif dist < self.chargeDistMax or dist > self.throwDistMax or self._level().isPointNearEdge(ourPos,self._running):
                if self._mode != 'charge':
                    self._mode = 'charge'
                    self._leadAmount = 0.01
                    self._chargeClosingIn = True
                    self._lastChargeDist = dist

            # we're too close to throw but too far to charge - either run away or just chill if we're near an edge
            elif dist < self.throwDistMin:
                # charge if either we're within charge range or cant retreat to throw
                self._mode = 'flee'

            # do some awesome jumps if we're running
            if self._running and dist > 1.2 and dist < 2.2 and bs.getGameTime()-self._lastJumpTime > 1000:
                self._lastJumpTime = bs.getGameTime()
                bs.nodeMessage(self.node,'jump')
                bs.gameTimer(500,bs.WeakCall(bs.nodeMessage,self.node,'jumpRelease'))
                
            # throw punches when real close
            if dist < (1.6 if self._running else 1.2) and canAttack:
                if random.random() < self.punchiness:
                    bs.nodeMessage(self.node,AttackMessage('punch'))
                    self.onPunchRelease()

    def __superHandleMessage(self,m):
        super(SpazBot,self).handleMessage(m)

    def onPunched(self,damage):
        activity = self._activity()
        if activity: activity.handleMessage(SpazBotPunchedMessage(self,damage))
        
    def handleMessage(self,m):

        # keep track of if we're being held and by who most recently
        if isinstance(m,bs.PickedUpMessage):
            self.__superHandleMessage(m) # augment standard behavior
            self.heldCount += 1
            pickedUpBy = m.node.sourcePlayer
            if pickedUpBy != -1:
                self.lastPlayerHeldByID = pickedUpBy

        elif isinstance(m,bs.DroppedMessage):
            self.__superHandleMessage(m) # augment standard behavior
            self.heldCount -= 1
            if self.heldCount < 0:
                print "ERROR: spaz heldCount < 0"
            # let's count someone dropping us as an attack..
            try: pickedUpBy = m.node.sourcePlayer
            except Exception: pickedUpBy = -1

            if pickedUpBy != -1:
                self.lastAttackerID = pickedUpBy
                self.lastAttackedTime = bs.getGameTime()
                self.lastAttackedType = ('pickedUp','default')
            
        elif isinstance(m,bs.DieMessage):

            # report deaths for scoring purposes
            if not self.dead:
                # if this guy was being held at the time of death, the holder is the killer
                if self.heldCount > 0 and self.lastPlayerHeldByID is not None:
                    killer = self.lastPlayerHeldByID
                else:
                    # otherwise if they were attacked by someone in the last few seconds
                    # that person's the killer.. otherwise it was a suicide
                    if self.lastAttackerID != None and bs.getGameTime() - self.lastAttackedTime < 4000:
                        killer = self.lastAttackerID
                    else:
                        killer = None
                activity = self._activity()
                if activity: activity.handleMessage(SpazBotDeathMessage(self,killer,m.how))
            self.__superHandleMessage(m) # augment standard behavior

        # keep track of the player who last hit us for point rewarding
        elif isinstance(m,bs.HitMessage):
            if m.source not in [None]:
                self.lastAttackerID = m.source
                self.lastAttackedTime = bs.getGameTime()
                self.lastAttackedType = (m.hitType,m.hitSubType)
            self.__superHandleMessage(m)
        else:
            Spaz.handleMessage(self,m)

class BomberBot(SpazBot):
    character='Spaz'
    punchiness=0.3

class BomberBotLame(BomberBot):
    color=gLameBotColor
    highlight=gLameBotHighlight
    punchiness = 0.2
    throwRate = 0.7
    throwiness = 0.1
    chargeSpeedMin = 0.6
    chargeSpeedMax = 0.6

class BomberBotStaticLame(BomberBotLame):
    static = True
    throwDistMin = 0.0

class BomberBotStatic(BomberBot):
    static = True
    throwDistMin = 0.0


class BomberBotPro(BomberBot):
    pointsMult = 2
    color=gProBotColor
    highlight = gProBotHighlight
    defaultBombCount = 3
    defaultBoxingGloves = True
    punchiness = 0.7
    throwRate = 1.3
    run = True
    runDistMin = 6.0

class BomberBotProShielded(BomberBotPro):
    pointsMult = 3
    defaultShields = True

class BomberBotProStatic(BomberBotPro):
    static = True
    throwDistMin = 0.0

class BomberBotProStaticShielded(BomberBotProShielded):
    static = True
    throwDistMin = 0.0

class ToughGuyBot(SpazBot):
    character = 'Kronk'
    punchiness = 0.9
    chargeDistMax = 9999.0
    chargeSpeedMin = 1.0
    chargeSpeedMax = 1.0
    throwDistMin = 9999
    throwDistMax = 9999

class ToughGuyBotLame(ToughGuyBot):
    color=gLameBotColor
    highlight=gLameBotHighlight
    punchiness = 0.3
    chargeSpeedMin = 0.6
    chargeSpeedMax = 0.6

class ToughGuyBotPro(ToughGuyBot):
    color=gProBotColor
    highlight=gProBotHighlight
    run = True
    runDistMin = 4.0
    defaultBoxingGloves = True
    punchiness = 0.95
    pointsMult = 2

class ToughGuyBotProShielded(ToughGuyBotPro):
    defaultShields = True
    pointsMult = 3

class NinjaBot(SpazBot):
    character = 'Snake Shadow'
    punchiness = 1.0
    run = True
    chargeDistMin = 10.0
    chargeDistMax = 9999.0
    chargeSpeedMin = 1.0
    chargeSpeedMax = 1.0
    throwDistMin = 9999
    throwDistMax = 9999
    pointsMult = 2

class NinjaBotPro(NinjaBot):
    color=gProBotColor
    highlight=gProBotHighlight
    defaultShields = True
    defaultBoxingGloves = True
    pointsMult = 3

class NinjaBotProShielded(NinjaBotPro):
    defaultShields = True
    pointsMult = 4

class ChickBot(SpazBot):
    character = 'Zoe'
    punchiness = 0.75
    throwiness = 0.7
    chargeDistMax = 1.0
    chargeSpeedMin = 0.3
    chargeSpeedMax = 0.5
    throwDistMin = 3.5
    throwDistMax = 5.5
    defaultBombType = 'impact'
    pointsMult = 2

class ChickBotStatic(ChickBot):
    static = True
    throwDistMin = 0.0

class ChickBotPro(ChickBot):
    color=gProBotColor
    highlight=gProBotHighlight
    defaultBombCount = 3
    defaultBoxingGloves = True
    chargeSpeedMin = 1.0
    chargeSpeedMax = 1.0
    punchiness = 0.9
    throwRate = 1.3
    run = True
    runDistMin = 6.0
    pointsMult = 3

class ChickBotProShielded(ChickBotPro):
    defaultShields = True
    pointsMult = 4

class MelBot(SpazBot):
    character = 'Mel'
    punchiness = 0.9
    throwiness = 1.0
    run = True
    chargeDistMin = 4.0
    chargeDistMax = 10.0
    chargeSpeedMin = 1.0
    chargeSpeedMax = 1.0
    throwDistMin = 0.0
    throwDistMax = 4.0
    throwRate = 2.0
    defaultBombType = 'sticky'
    defaultBombCount = 3
    pointsMult = 3

class MelBotStatic(MelBot):
    static = True

class PirateBot(SpazBot):
    character = 'Jack Morgan'
    run = True
    chargeDistMin = 0.0
    chargeDistMax = 9999
    chargeSpeedMin = 1.0
    chargeSpeedMax = 1.0
    throwDistMin = 9999
    throwDistMax = 9999
    startCursed = True
    pointsMult = 4

class PirateBotNoTimeLimit(PirateBot):
    curseTime = -1

class PirateBotShielded(PirateBot):
    defaultShields = True
    pointsMult = 5
    
class BotSet(object):
    def __init__(self,players):
        # we spread our bots out over a few lists so we can update them in a staggered fashion
        self._botListCount = 5
        self._botAddList = 0
        self._botUpdateList = 0
        self._botLists = [[] for i in range(self._botListCount)]
        self._botUpdateTimer = bs.Timer(50,bs.WeakCall(self._update),repeat=True)
        self._players = players

    def __del__(self):
        self.clear()

    def haveLivingBots(self):
        return any((any((not a.dead for a in l)) for l in self._botLists))

    def getLivingBots(self):
        bots = []
        for l in self._botLists:
            for b in l:
                if not b.dead: bots.append(b)
        return bots

    def _update(self):

        # update one of our bot lists each time through..
        # first off, remove dead bots from the list
        # (we check exists() here instead of dead.. we want to keep them around even if they're just a corpse)
        botList = self._botLists[self._botUpdateList] = [b for b in self._botLists[self._botUpdateList] if b.exists()]
        self._botUpdateList = (self._botUpdateList+1)%self._botListCount

        # update our list of player points for the bots to use
        playerPts = []
        for player in self._players.values():
            try:
                if not player.actor.dead:
                    playerPts.append((bsVector.Vector(*player.actor.node.translate),
                                     bsVector.Vector(*player.actor.node.velocity)))
            except Exception,e: pass

        for b in botList:
            b.setPlayerPts(playerPts)
            b.updateAI()

    def clear(self):
        for i in range(len(self._botLists)):
            for b in self._botLists[i]:
                b.handleMessage(bs.DieMessage(immediate=True))
            self._botLists[i] = []
        
    def celebrate(self,duration):
        for l in self._botLists:
            for b in l:
                if b.node.exists():
                    bs.nodeMessage(b.node,'celebrate',duration)

    def stopMoving(self):
        self._botUpdateTimer = None
        for l in self._botLists:
            for b in l:
                if b.node.exists():
                    bs.nodeMessage(b.node,'moveLeftRight',0)
                    bs.nodeMessage(b.node,'moveUpDown',0)
        
    def finalCelebrate(self):
        self._botUpdateTimer = None
        # at this point stop doing anything but jumping and celebrating
        for l in self._botLists:
            for b in l:
                if b.node.exists():
                    bs.nodeMessage(b.node,'moveLeftRight',0)
                    bs.nodeMessage(b.node,'moveUpDown',0)
                    bs.gameTimer(random.randrange(0,500),bs.WeakCall(bs.nodeMessage,b.node,'celebrate',10000))
                    jumpDuration = random.randrange(400,500)
                    j = random.randrange(0,200)
                    for i in range(10):
                        bs.gameTimer(j,bs.WeakCall(bs.nodeMessage,b.node,'jump'))
                        bs.gameTimer(j+30,bs.WeakCall(bs.nodeMessage,b.node,'jumpRelease'))
                        j += jumpDuration
                    bs.gameTimer(random.randrange(0,1000),bs.WeakCall(bs.nodeMessage,b.node,'attackSound'))
                    bs.gameTimer(random.randrange(1000,2000),bs.WeakCall(bs.nodeMessage,b.node,'attackSound'))
                    bs.gameTimer(random.randrange(2000,3000),bs.WeakCall(bs.nodeMessage,b.node,'attackSound'))

    def addBot(self,bot):
        self._botLists[self._botAddList].append(bot)
        self._botAddList = (self._botAddList+1)%self._botListCount


# define our built-in characters...

###############  SPAZ   ##################
t = Appearance("Spaz")

t.colorTexture = "neoSpazColor"
t.colorMaskTexture = "neoSpazColorMask"

t.iconTexture = "neoSpazIcon"
t.iconMaskTexture = "neoSpazIconColorMask"

t.headModel = "neoSpazHead"
t.torsoModel = "neoSpazTorso"
t.pelvisModel = "neoSpazPelvis"
t.upperArmModel = "neoSpazUpperArm"
t.foreArmModel = "neoSpazForeArm"
t.handModel = "neoSpazHand"
t.upperLegModel = "neoSpazUpperLeg"
t.lowerLegModel = "neoSpazLowerLeg"
t.toesModel = "neoSpazToes"

t.jumpSounds=["spazJump01",
              "spazJump02",
              "spazJump03",
              "spazJump04"]
t.attackSounds=["spazAttack01",
                "spazAttack02",
                "spazAttack03",
                "spazAttack04"]
t.impactSounds=["spazImpact01",
                "spazImpact02",
                "spazImpact03",
                "spazImpact04"]
t.deathSounds=["spazDeath01"]
t.pickupSounds=["spazPickup01"]
t.fallSounds=["spazFall01"]

t.style = 'spaz'


###############  Zoe   ##################
t = Appearance("Zoe")

t.colorTexture = "zoeColor"
t.colorMaskTexture = "zoeColorMask"

t.iconTexture = "zoeIcon"
t.iconMaskTexture = "zoeIconColorMask"

t.headModel = "zoeHead"
t.torsoModel = "zoeTorso"
t.pelvisModel = "zoePelvis"
t.upperArmModel = "zoeUpperArm"
t.foreArmModel = "zoeForeArm"
t.handModel = "zoeHand"
t.upperLegModel = "zoeUpperLeg"
t.lowerLegModel = "zoeLowerLeg"
t.toesModel = "zoeToes"

t.jumpSounds=["zoeJump01",
              "zoeJump02",
              "zoeJump03"]
t.attackSounds=["zoeAttack01",
                "zoeAttack02",
                "zoeAttack03",
                "zoeAttack04"]
t.impactSounds=["zoeImpact01",
                "zoeImpact02",
                "zoeImpact03",
                "zoeImpact04"]
t.deathSounds=["zoeDeath01"]
t.pickupSounds=["zoePickup01"]
t.fallSounds=["zoeFall01"]

t.style = 'female'

###############  Ninja   ##################
t = Appearance("Snake Shadow")

t.colorTexture = "ninjaColor"
t.colorMaskTexture = "ninjaColorMask"

t.iconTexture = "ninjaIcon"
t.iconMaskTexture = "ninjaIconColorMask"

t.headModel = "ninjaHead"
t.torsoModel = "ninjaTorso"
t.pelvisModel = "ninjaPelvis"
t.upperArmModel = "ninjaUpperArm"
t.foreArmModel = "ninjaForeArm"
t.handModel = "ninjaHand"
t.upperLegModel = "ninjaUpperLeg"
t.lowerLegModel = "ninjaLowerLeg"
t.toesModel = "ninjaToes"

ninjaAttacks = ['ninjaAttack'+str(i+1)+'' for i in range(7)]
ninjaHits = ['ninjaHit'+str(i+1)+'' for i in range(8)]
ninjaJumps = ['ninjaAttack'+str(i+1)+'' for i in range(7)]

t.jumpSounds=ninjaJumps
t.attackSounds=ninjaAttacks
t.impactSounds=ninjaHits
t.deathSounds=["ninjaDeath1"]
t.pickupSounds=ninjaAttacks
t.fallSounds=["ninjaFall1"]

t.style = 'ninja'


###############  Kronk   ##################
t = Appearance("Kronk")

t.colorTexture = "kronk"
t.colorMaskTexture = "kronkColorMask"

t.iconTexture = "kronkIcon"
t.iconMaskTexture = "kronkIconColorMask"

t.headModel = "kronkHead"
t.torsoModel = "kronkTorso"
t.pelvisModel = "kronkPelvis"
t.upperArmModel = "kronkUpperArm"
t.foreArmModel = "kronkForeArm"
t.handModel = "kronkHand"
t.upperLegModel = "kronkUpperLeg"
t.lowerLegModel = "kronkLowerLeg"
t.toesModel = "kronkToes"

kronkSounds = ["kronk1",
              "kronk2",
              "kronk3",
              "kronk4",
              "kronk5",
              "kronk6",
              "kronk7",
              "kronk8",
              "kronk9",
              "kronk10"]
t.jumpSounds=kronkSounds
t.attackSounds=kronkSounds
t.impactSounds=kronkSounds
t.deathSounds=["kronkDeath"]
t.pickupSounds=kronkSounds
t.fallSounds=["kronkFall"]

t.style = 'kronk'


###############  MEL   ##################
t = Appearance("Mel")

t.colorTexture = "melColor"
t.colorMaskTexture = "melColorMask"

t.iconTexture = "melIcon"
t.iconMaskTexture = "melIconColorMask"

t.headModel =     "melHead"
t.torsoModel =    "melTorso"
t.pelvisModel = "kronkPelvis"
t.upperArmModel = "melUpperArm"
t.foreArmModel =  "melForeArm"
t.handModel =     "melHand"
t.upperLegModel = "melUpperLeg"
t.lowerLegModel = "melLowerLeg"
t.toesModel =     "melToes"

melSounds = ["mel01",
              "mel02",
              "mel03",
              "mel04",
              "mel05",
              "mel06",
              "mel07",
              "mel08",
              "mel09",
              "mel10"]

t.attackSounds = melSounds
t.jumpSounds = melSounds
t.impactSounds = melSounds
t.deathSounds=["melDeath01"]
t.pickupSounds = melSounds
t.fallSounds=["melFall01"]

t.style = 'mel'


###############  Jack Morgan   ##################

t = Appearance("Jack Morgan")

t.colorTexture = "jackColor"
t.colorMaskTexture = "jackColorMask"

t.iconTexture = "jackIcon"
t.iconMaskTexture = "jackIconColorMask"

t.headModel =     "jackHead"
t.torsoModel =    "jackTorso"
t.pelvisModel = "kronkPelvis"
t.upperArmModel = "jackUpperArm"
t.foreArmModel =  "jackForeArm"
t.handModel =     "jackHand"
t.upperLegModel = "jackUpperLeg"
t.lowerLegModel = "jackLowerLeg"
t.toesModel =     "jackToes"

hitSounds = ["jackHit01",
             "jackHit02",
             "jackHit03",
             "jackHit04",
             "jackHit05",
             "jackHit06",
             "jackHit07"]

sounds = ["jack01",
          "jack02",
          "jack03",
          "jack04",
          "jack05",
          "jack06"]

t.attackSounds = sounds
t.jumpSounds = sounds
t.impactSounds = hitSounds
t.deathSounds=["jackDeath01"]
t.pickupSounds = sounds
t.fallSounds=["jackFall01"]

t.style = 'pirate'


###############  SANTA   ##################

t = Appearance("Santa Claus")

t.colorTexture = "santaColor"
t.colorMaskTexture = "santaColorMask"

t.iconTexture = "santaIcon"
t.iconMaskTexture = "santaIconColorMask"

t.headModel =     "santaHead"
t.torsoModel =    "santaTorso"
t.pelvisModel = "kronkPelvis"
t.upperArmModel = "santaUpperArm"
t.foreArmModel =  "santaForeArm"
t.handModel =     "santaHand"
t.upperLegModel = "santaUpperLeg"
t.lowerLegModel = "santaLowerLeg"
t.toesModel =     "santaToes"

hitSounds = ['santaHit01','santaHit02','santaHit03','santaHit04']
sounds = ['santa01','santa02','santa03','santa04','santa05']

t.attackSounds = sounds
t.jumpSounds = sounds
t.impactSounds = hitSounds
t.deathSounds=["santaDeath"]
t.pickupSounds = sounds
t.fallSounds=["santaFall"]

t.style = 'santa'

###############  FROSTY   ##################

t = Appearance("Frosty")

t.colorTexture = "frostyColor"
t.colorMaskTexture = "frostyColorMask"

t.iconTexture = "frostyIcon"
t.iconMaskTexture = "frostyIconColorMask"

t.headModel =     "frostyHead"
t.torsoModel =    "frostyTorso"
t.pelvisModel = "frostyPelvis"
t.upperArmModel = "frostyUpperArm"
t.foreArmModel =  "frostyForeArm"
t.handModel =     "frostyHand"
t.upperLegModel = "frostyUpperLeg"
t.lowerLegModel = "frostyLowerLeg"
t.toesModel =     "frostyToes"

frostySounds = ['frosty01','frosty02','frosty03','frosty04','frosty05']
frostyHitSounds = ['frostyHit01','frostyHit02','frostyHit03']

t.attackSounds = frostySounds
t.jumpSounds = frostySounds
t.impactSounds = frostyHitSounds
t.deathSounds=["frostyDeath"]
t.pickupSounds = frostySounds
t.fallSounds=["frostyFall"]

t.style = 'frosty'
